
/*
 *    miracleshaper - Traffic shaper and prioritizer for DSLs.
 *    Copyright (C) 2007  Florian Zumbiehl <florz@florz.de>
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <linux/if.h>
#include <linux/if_tun.h>
#include <netpacket/packet.h>
#include <net/ethernet.h>
#include <netinet/in.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>
#include <stdbool.h>
#include <stdlib.h>
#include <sched.h>
#include <stdio.h>
#include <errno.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <syslog.h>

#include "config.h"
#include "utils.h"
#include "event.h"
#include "prioqueue.h"
#include "pkt.h"

#ifndef TUNSETMARK
#define TUNSETMARK    _IOW('T', 206, int)
#endif

static struct prioqueue *downstreamq,*upstreamq;
static int rawsock,taps[FRONTEND_INTERFACE_COUNT];

static int create_tap(const char *name,long long mac,bool persistent){
	int fd,sock;
	struct ifreq ifr;

	if((fd=open("/dev/net/tun",O_RDWR))<0)die(true,"open(\"/dev/net/tun\")");
	strcpy(ifr.ifr_name,name);
	ifr.ifr_flags=IFF_TAP|IFF_NO_PI;
	if(ioctl(fd,TUNSETIFF,(void *)&ifr)<0)die(true,"ioctl(%s,TUNSETIFF)",name);
	if((sock=socket(PF_UNIX,SOCK_STREAM,0))<0)die(true,"socket()");
	if(ioctl(sock,SIOCGIFHWADDR,(void *)&ifr)<0)die(true,"ioctl(%s,SIOCGIFHWADDR)",name);
	if(ifr.ifr_hwaddr.sa_family!=ARPHRD_ETHER)die(false,"%s is not an ethernet interface?!",name);
	if(mac_from_6oct(ifr.ifr_hwaddr.sa_data)!=mac){
		if(ioctl(sock,SIOCGIFFLAGS,(void *)&ifr)<0)die(true,"ioctl(%s,SIOCGIFFLAGS)",name);
		ifr.ifr_flags&=~IFF_UP;
		if(ioctl(sock,SIOCSIFFLAGS,(void *)&ifr)<0)die(true,"ioctl(%s,SIOCSIFFLAGS)",name);
		ifr.ifr_hwaddr.sa_family=ARPHRD_ETHER;
		mac_to_6oct(ifr.ifr_hwaddr.sa_data,mac);
		if(ioctl(sock,SIOCSIFHWADDR,(void *)&ifr)<0)die(true,"ioctl(%s,SIOCSIFHWADDR)",name);
	}
	if(ioctl(sock,SIOCGIFFLAGS,(void *)&ifr)<0)die(true,"ioctl(%s,SIOCGIFFLAGS)",name);
	ifr.ifr_flags|=IFF_UP;
	if(ioctl(sock,SIOCSIFFLAGS,(void *)&ifr)<0)die(true,"ioctl(%s,SIOCSIFFLAGS)",name);
	if(close(sock)<0)die(true,"close(sock)");
	if(ioctl(fd,TUNSETPERSIST,persistent?1:0)<0)die(true,"ioctl(%s,TUNSETPERSIST)",name);
#ifdef USEMARK
	if(ioctl(fd,TUNSETMARK,1)<0)die(true,"ioctl(%s,TUNSETMARK)",name);
#else
	ioctl(fd,TUNSETMARK,0);
#endif
	return fd;
}

static int profile_flags;

static inline void profile_activate(int,int);

#include "classifiers.c"

static int profile_timers[PROFILE_COUNT];

static inline void profile_activate(int profile,int timeout){
	if(!(profile_flags&(1<<profile))){
		profile_flags|=1<<profile;
		profile_commit();
		event_timer_enable(profile_timers[profile]);
	}
	event_timer_set_delay(profile_timers[profile],timeout*event_timer_hz);
}

static void profile_deactivate(void *arg){
	int *profilep=arg;
	int profile;

	profile=profilep-profile_timers;
	profile_flags&=~(1<<profile);
	profile_commit();
	event_timer_disable(*profilep);
}

static ssize_t rawsock_write(struct pkt *pkt){
	return write(rawsock,pkt_get_buf(pkt),pkt_get_len(pkt));
}

static ssize_t tap_write(struct pkt *pkt){
	long long mac;

	pkt_pinit(pkt);
	if((mac=pkt_eth_get_dst(pkt))>=0){
		if(mac>=FRONTEND_INTERFACE_BASE_MAC&&mac<=FRONTEND_INTERFACE_BASE_MAC+FRONTEND_INTERFACE_COUNT-1)
			return write(taps[mac-FRONTEND_INTERFACE_BASE_MAC],pkt_get_buf(pkt),pkt_get_len(pkt));
		if(mac==0xffffff){
			ssize_t r1,r2=0;

			for(int x=0;x<FRONTEND_INTERFACE_COUNT;x++){
				r1=write(taps[x],pkt_get_buf(pkt),pkt_get_len(pkt));
				if(r1<r2)r2=r1;
			}
			return r2;
		}
	}
	return pkt_get_len(pkt);
}

int main(int argc,char **argv){
	int optchar;
	bool daemonize=true;

	openlog("mshaper",LOG_CONS|LOG_NDELAY|LOG_PERROR|LOG_PID,LOG_DAEMON);
	syslog(LOG_INFO,"version " VERSION " starting");

	opterr=0;
	while((optchar=getopt(argc,argv,"hdr:"))>0){
		switch(optchar){
			case 'h':
				printf("usage: %s [ -h | -d | -r <ifacename> ]\n",argv[0]);
				printf("\n");
				printf("  -h: help\n");
				printf("  -d: don't detach from terminal\n");
				printf("  -r: remove specified persistent tap interface\n");
				printf("\n");
				exit(0);
			case 'd':
				daemonize=false;
				break;
			case 'r':
				syslog(LOG_INFO,"removing tap interface: %s",optarg);
				if(strlen(optarg)>IFNAMSIZ-1)die(false,"interface name too long: %s",optarg);
				create_tap(optarg,FRONTEND_INTERFACE_BASE_MAC,false);
				exit(0);
			case '?':
				die(false,"Unknown command line option '%c'",optopt);
				break;
		}
	}
	if(argv[optind])die(false,"unexpected command line arguments");

	if(daemonize){
		int fd;

		switch(fork()){
			case -1:
				die(true,"fork()");
			default:
				exit(0);
			case 0:
				;
		}
		if(setsid()<0)die(true,"setsid()");
		switch(fork()){
			case -1:
				die(true,"fork()");
			default:
				exit(0);
			case 0:
				;
		}
		fd=open("/dev/null",O_RDWR);
		if(dup2(fd,0)<0)die(true,"dup2()");
		if(dup2(fd,1)<0)die(true,"dup2()");
		if(dup2(fd,2)<0)die(true,"dup2()");
		if(close(fd)<0)die(true,"close()");
	}
	if(chdir("/")<0)die(true,"chdir(\"/\")");

	event_init();

	{
		struct ifreq ifr;
		struct sockaddr_ll rawsock_addr;

		if((rawsock=socket(PF_PACKET,SOCK_RAW,htons(ETH_P_ALL)))<0)die(true,"socket(rawsock)");
		memset(&rawsock_addr,0,sizeof(rawsock_addr));
		rawsock_addr.sll_family=AF_PACKET;
		strcpy(ifr.ifr_name,BACKEND_INTERFACE);
		if(ioctl(rawsock,SIOCGIFINDEX,&ifr)<0)die(true,"ioctl(rawsock,SIOCGIFINDEX)");
		rawsock_addr.sll_ifindex=ifr.ifr_ifindex;
		if(bind(rawsock,(struct sockaddr *)&rawsock_addr,sizeof(rawsock_addr))<0)die(true,"bind(rawsock)");

		strcpy(ifr.ifr_name,BACKEND_INTERFACE);
		if(ioctl(rawsock,SIOCGIFFLAGS,&ifr)<0)die(true,"ioctl(rawsock,SIOCGIFFLAGS)");
		ifr.ifr_flags|=IFF_PROMISC;
		if(ioctl(rawsock,SIOCSIFFLAGS,&ifr)<0)die(true,"ioctl(rawsock,SIOCSIFFLAGS)");

	}

	upstreamq=prioqueue_create(
			"upstream",
			UPSTREAM_QUEUE_MAX_PACKETS,
			UPSTREAM_BANDWIDTH,
			upstream_classify,
			UPSTREAM_MAX_PRIORITY,
			rawsock_write);

	for(int x=0;x<FRONTEND_INTERFACE_COUNT;x++){
		char buf[IFNAMSIZ];

		if(snprintf(buf,sizeof(buf),"%s%i",FRONTEND_INTERFACE_PREFIX,x)>=sizeof(buf))
			die(false,"interface name too long: %s%i",FRONTEND_INTERFACE_PREFIX,x);
		taps[x]=create_tap(buf,FRONTEND_INTERFACE_BASE_MAC+x,true);
		event_add_readfd(taps[x],prioqueue_read_cb,upstreamq);
	}

	downstreamq=prioqueue_create(
			"downstream",
			DOWNSTREAM_QUEUE_MAX_PACKETS,
			DOWNSTREAM_BANDWIDTH,
			downstream_classify,
			DOWNSTRAM_MAX_PRIORITITY,
			tap_write);

	event_add_readfd(rawsock,prioqueue_read_cb,downstreamq);

	for(int x=PROFILE_COUNT;x--;)
		profile_timers[x]=event_timer_create(profile_deactivate,profile_timers+x);

	{
		struct sched_param schp;

		schp.sched_priority=1;
		if(sched_setscheduler(0,SCHED_FIFO,&schp)<0)die(true,"sched_setscheduler()");
	}

	syslog(LOG_INFO,"initialisation complete");
	event_dispatch();
}

