
/*
 *    miracleshaper - Traffic shaper and prioritizer for DSLs.
 *    Copyright (C) 2007  Florian Zumbiehl <florz@florz.de>
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

#include <stdlib.h>
#include <syslog.h>
#include <stdio.h>

#include "config.h"
#include "utils.h"
#include "event.h"
#include "prioqueue.h"

#define MTU 1514
#define MTU_ALIGNMENT (__alignof__(char __attribute__ ((aligned))))
#define MTU_ALIGNED (MTU+MTU_ALIGNMENT-1)/MTU_ALIGNMENT*MTU_ALIGNMENT

/*
 * As it might be somewhat confusing if you just try to understand it from the
 * source, let me explain how the "ECN queueing" works. That is: What happens
 * if we enqueue a packet that causes the queue to grow larger that its
 * current size limit?
 *
 * Well, as long as the queue is larger than the current limit allows, we
 * fetch the respective last packet that was enqueued for the lowest priority
 * that we currently do have packets for. Then, if that packet does not
 * indicate an ECN-capable transport, it's simply dropped. If it _does_
 * indicate an ECN-capable transport, though, we do ECN-mark it and then
 * put it into a second "overflow queue" that is there only for holding
 * packets that don't fit into the main queue but have been ECN-marked instead
 * of dropped. Whenever the main queue shrinks far enough, this overflow queue
 * is dequeued into the main queue again.
 *
 * As we have to retain the classification of packets when moving them from
 * one queue to the other, you may assume for now that we do have two
 * full-blown queues with the configured number of priority slots each. Then
 * we can simply put packets into the slot with the same number as the one
 * it's coming from in the respective other queue.
 *
 * If you do think about it for a moment, though, you will realize that
 * there are never packets for the same slot number on both queues, except
 * for the lowest priority slot that there are currently packets in the
 * main queue for. So we actually do need only one additional slot in
 * the single queue structure and some index that stores the slot that is
 * the current boundary between the main and the overflow queue.
 *
 * I hope that this suffices for you to understand the details from the
 * source.
 */

struct prioqueue {
	const char *name;
	int num_priorities,tx_prio,ecn_boundary,data_rate;
	ssize_t (*write_func)(struct pkt *);
	int timer;
	int (*classifier)(struct pkt *);
	int max_queue_len;
	long max_queue_size,queue_size,ecn_queue_size;
	struct list pool,queues[0];
};

static inline int packetbytes_to_linkbits(int size){
	//size+=2;		// PPP "header"
	//size+=6;		// PPPoE header
	//size+=6+6+2;		// ethernet header
	//size+=4;		// ethernet FCS
	size+=3+3+2+2;		// RFC2684 802.3 bridged LLC
	size+=1+1+2+4;		// AAL5 trailer
	size=(size+48-1)/48;	// round up to full ATM cell size
	size*=53;		// ATM cell header
	size*=8;		// convert to bits
	return size;
}

static void prioqueue_write_pkt(struct prioqueue *q,struct pkt *pkt){
	if(q->write_func(pkt)<0)die(true,"send(%s)",q->name);
	event_timer_advance(q->timer,packetbytes_to_linkbits(pkt_get_len(pkt))*event_timer_hz/q->data_rate);
}

static void prioqueue_try_empty_ecn_queue(struct prioqueue *q){
	int *x;
	struct list *pkt_li;
	struct pkt *pkt;

	x=&q->ecn_boundary;
	while(*x<q->num_priorities&&q->queue_size<q->max_queue_size){
		pkt_li=list_dequeue(q->queues+*x+1);
		pkt=listitem_get_data_p(pkt_li);
		if(packetbytes_to_linkbits(pkt_get_len(pkt))<=q->max_queue_size-q->queue_size){
			q->ecn_queue_size-=packetbytes_to_linkbits(pkt_get_len(pkt));
			list_enqueue(q->queues+*x,pkt_li);
			q->queue_size+=packetbytes_to_linkbits(pkt_get_len(pkt));
		}else{
			list_prepend(q->queues+*x+1,pkt_li);
			break;
		}
		while(*x<q->num_priorities&&list_empty(q->queues+*x+1))(*x)++;
	}
}

static void prioqueue_drop_overflow(struct prioqueue *q){
	int x;

	x=q->ecn_boundary;
	while(q->queue_size>q->max_queue_size){
		struct list *drop_pkt_li;
		struct pkt *drop_pkt;

		while(list_empty(q->queues+x))x--;
		drop_pkt_li=list_fetchlast(q->queues+x);
		drop_pkt=listitem_get_data_p(drop_pkt_li);
		q->queue_size-=packetbytes_to_linkbits(pkt_get_len(drop_pkt));
		if(pkt_ip_mark_ecn_x(drop_pkt)){
			syslog(LOG_DEBUG,"ECN-marking packet on %s\n",q->name);
			list_prepend(q->queues+x+1,drop_pkt_li);
			q->ecn_queue_size+=packetbytes_to_linkbits(pkt_get_len(drop_pkt));
			q->ecn_boundary=x;
		}else{
			syslog(LOG_DEBUG,"dropping packet on %s\n",q->name);
			list_prepend(&q->pool,drop_pkt_li);
		}
	}
}

static void prioqueue_write(void *arg){
	struct prioqueue *q=arg;
	struct list *pkt_li;
	struct pkt *pkt;

	pkt_li=list_dequeue(q->queues+q->tx_prio);
	pkt=listitem_get_data_p(pkt_li);
	q->queue_size-=packetbytes_to_linkbits(pkt_get_len(pkt));
	prioqueue_write_pkt(q,pkt);
	list_prepend(&q->pool,pkt_li);
	prioqueue_try_empty_ecn_queue(q);
	while(list_empty(q->queues+q->tx_prio)){
		if(q->tx_prio==q->num_priorities-1){
			event_timer_disable(q->timer);
			break;
		}
		q->tx_prio++;
	}
}

static void prioqueue_read(struct prioqueue *q,int fd){
	struct list *pkt_li;
	struct pkt *pkt;
#ifdef USEMARK
	struct iovec iov[2];
	int iovl;
	ssize_t readr;
	uint32_t mark;
#endif

	if(list_empty(&q->pool)){
		for(int x=q->num_priorities+1;x--;){
			if(!list_empty(q->queues+x)){
				struct list *drop_pkt;

				syslog(LOG_NOTICE,"dropping packet on %s (too few buffers available)",q->name);
				drop_pkt=list_fetchlast(q->queues+x);
				if(x<q->ecn_boundary){
					q->queue_size-=packetbytes_to_linkbits(pkt_get_len(listitem_get_data_p(drop_pkt)));
				}else{
					q->ecn_queue_size-=packetbytes_to_linkbits(pkt_get_len(listitem_get_data_p(drop_pkt)));
				}
				list_prepend(&q->pool,drop_pkt);
				if(x==q->ecn_boundary+1&&list_empty(q->queues+x))q->ecn_boundary=q->num_priorities;
				break;
			}
		}
	}
	pkt_li=list_dequeue(&q->pool);
	pkt=listitem_get_data_p(pkt_li);
#ifdef USEMARK
	if(!strcmp(q->name,"upstream")){
		iov[0].iov_base=&mark;
		iov[0].iov_len=sizeof(mark);
		iov[1].iov_base=pkt_get_buf(pkt);
		iov[1].iov_len=MTU;
		iovl=2;
	}else{
		iov[0].iov_base=pkt_get_buf(pkt);
		iov[0].iov_len=MTU;
		iovl=1;
	}
	if((readr=readv(fd,iov,iovl))<0){
#else
	if(pkt_set_len(pkt,read(fd,pkt_get_buf(pkt),MTU))<0){
#endif
		die(true,"recv(%s)",q->name);
	}else{
		int prio;

#ifdef USEMARK
		if(!strcmp(q->name,"upstream")){
			mark=ntohl(mark);
			printf("MARK: %i\n",mark);
			pkt_set_mark(pkt,mark);
			pkt_set_len(pkt,readr-sizeof(mark));
		}else{
			pkt_set_len(pkt,readr);
		}
#endif
		if((prio=q->classifier(pkt))<0){
			if(prio==PRIO_IMMEDIATE)prioqueue_write_pkt(q,pkt);
			list_prepend(&q->pool,pkt_li);
		}else{
			if(prio<q->ecn_boundary){
				list_enqueue(q->queues+prio,pkt_li);
				q->queue_size+=packetbytes_to_linkbits(pkt_get_len(pkt));
				prioqueue_drop_overflow(q);
			}else{
				if(pkt_ip_mark_ecn_x(pkt)){
					syslog(LOG_DEBUG,"ECN-marking packet on %s\n",q->name);
					list_enqueue(q->queues+prio+1,pkt_li);
					q->ecn_queue_size+=packetbytes_to_linkbits(pkt_get_len(pkt));
				}else{
					syslog(LOG_DEBUG,"dropping packet on %s\n",q->name);
					list_prepend(&q->pool,pkt_li);
				}
			}
			if(prio<q->tx_prio)q->tx_prio=prio;
			event_timer_enable(q->timer);
		}
	}
}

void prioqueue_read_cb(int fd,void *arg){
	prioqueue_read(arg,fd);
}

struct prioqueue *prioqueue_create(
		const char *name,
		int max_queue_len,
		int data_rate,
		int (*classifier)(struct pkt *),
		int max_priority,
		ssize_t (*write_func)(struct pkt *)){
	struct prioqueue *q;
	char *mem;

	if(!(q=malloc(sizeof(struct prioqueue)+sizeof(*q->queues)*(max_priority+1+1))))
		die(false,"malloc()ing prioqueue failed");
	q->name=name;
	q->max_queue_len=max_queue_len;
	list_init(&q->pool);
	if(!(mem=malloc((sizeof(struct list)+sizeof(struct pkt)+MTU_ALIGNED)*max_queue_len)))
		die(false,"malloc()ing queue buffers failed");
	for(int x=max_queue_len;x--;){
		list_enqueue(&q->pool,(struct list *)(mem+x*(sizeof(struct list)+sizeof(struct pkt)+MTU_ALIGNED)));
	q->data_rate=data_rate;
	q->classifier=classifier;
	q->num_priorities=max_priority+1;
	q->tx_prio=max_priority;
	q->ecn_boundary=max_priority+1;
	for(int x=q->num_priorities+1;x--;)
		list_init(q->queues+x);
	}
	q->write_func=write_func;
	q->timer=event_timer_create(prioqueue_write,q);
	prioqueue_set_max_queue_size(q,0);
	q->queue_size=0;
	q->ecn_queue_size=0;
	return q;
}

void prioqueue_set_data_rate(struct prioqueue *q,int rate){
	q->data_rate=rate;
}

void prioqueue_set_max_queue_size(struct prioqueue *q,int fullframes){
	long old;
	
	old=q->max_queue_size;
	q->max_queue_size=(long)(fullframes?fullframes:q->max_queue_len)*packetbytes_to_linkbits(MTU);
	if(q->max_queue_size<old){
		prioqueue_drop_overflow(q);
	}else if(q->max_queue_size>old){
		prioqueue_try_empty_ecn_queue(q);
	}
}

